(* 
 * Arithmetic constraints.
 * See developper doc first.
 *)
open Softcore

(* Kind of equality *)
type inequality = Le | Gt | Eq

  (* Various facilities:
   * - conversion to string
   * - reversing an inequality (x > y is equivalent to y <= x-1 and *not to* y <= x, but x = y is equivalent to y = x,
   *   so reverse_sgn gives the reverse sign and reverse_int gives the function to apply to the integer in order to get the equivalence :
   *   fun x -> x-1 if inequality, fun x -> x if equality)
   * - conversion to the corresponding functions of ocaml
   *)
let ineq_to_string = function Le -> " <= " | Gt -> " > "  | Eq -> " = "
let reverse_sgn    = function Le -> Gt   | Gt -> Le   | Eq -> Eq
let reverse_int    = function Le         | Gt -> pred | Eq -> ident
let to_fun ineq (x:int) (y:int) = match ineq with Le -> x <= y | Gt -> x > y | Eq -> x = y 
let ineq_equal ineq1 ineq2 =
  match ineq1,ineq2 with
  |Le,Le | Gt,Gt | Eq,Eq -> true
  |_                     -> false

(* Type of the constraints, either true, or false, or the comparison between a variable and an int
 * i.e. the only non-trivial constraints are of the form n {<=,>,=} k where n is a variable, k an integer
 * As a consequence, n < 4 & n > 0 is not a supported constraint !
 * *We don't need more for regular schemata*, the strategy implies that such constraints should not occur.
 *)
type t =
  |TrueConstraint
  |FalseConstraint
  |AtomConstraint of Variable.t * inequality * int

let equal c1 c2 =
  match c1,c2 with
  |TrueConstraint,TrueConstraint                           -> true
  |FalseConstraint,FalseConstraint                         -> true
  |AtomConstraint(v1,ineq1,n1),AtomConstraint(v2,ineq2,n2) -> n1 = n2 && ineq_equal ineq1 ineq2 && Variable.equal v1 v2
  |_                                                       -> false

(* Print stuff *)
let to_string ?(parenthesized=false) = 
  let paren s = if parenthesized then s else "" in
  function
    |TrueConstraint -> ""
    |FalseConstraint -> "(FALSE)"
    |AtomConstraint(v,ineq,n) -> (paren "(") ^ (Variable.to_string v) ^ (ineq_to_string ineq) ^ (string_of_int n) ^ (paren ")")
let print = print_endline <|| to_string

(* Exceptions *)
exception UnsupportedConstraint
exception NonRegularBounds

(* Selectors *)
let of_bool = function true -> TrueConstraint | false -> FalseConstraint
let is_false c = match c with FalseConstraint -> true | _ -> false
let is_trivial c = match c with FalseConstraint | TrueConstraint -> true | _ -> false
let is_equality = function AtomConstraint(_,Eq,_) -> true | _ -> false

(* Constructors *)
let top     = TrueConstraint

(* Two possibilities to create a constraint, see developper doc.
 * Roughly: only one of two bounds can contain a variable.
 * If regular_bounds is true then we don't authorize each term to have a variable.
 *)
let constraint_atom_internal ~regular_bounds t1 ineq t2 =
  match (Indexes.get_var t1),(Indexes.get_var t2) with
  |Some(v),None                              -> AtomConstraint(v, ineq, (Indexes.get_shift t2)-(Indexes.get_shift t1))
  |None, Some(v)                             -> AtomConstraint(v,reverse_sgn ineq, reverse_int ineq ((Indexes.get_shift t1)-(Indexes.get_shift t2)))
  |None,None        when regular_bounds      -> raise NonRegularBounds
  |Some(v),Some(v') when regular_bounds      -> raise NonRegularBounds
  |Some(v),Some(v') when Variable.equal v v' -> of_bool (to_fun ineq (Indexes.get_shift t1) (Indexes.get_shift t2))
  |None,None                                 -> of_bool (to_fun ineq (Indexes.get_shift t1) (Indexes.get_shift t2))
  |_                                         -> raise UnsupportedConstraint

let constraint_atom = constraint_atom_internal ~regular_bounds:false
let bounds_constraint = constraint_atom_internal ~regular_bounds:true 

(* Combines two constraints.
 * As previously said we don't support general conjunction of constraints.
 * So constraints can be combined only if one is trivial or if both deal with the same variable in some "convenient" way.
 *)
let bounds_conjunction c1 c2 =
  match c1,c2 with
  |TrueConstraint,c
  |c,TrueConstraint                                                                        -> c
  |FalseConstraint,_
  |_,FalseConstraint                                                                       -> FalseConstraint
  |AtomConstraint(v,Eq,n),AtomConstraint(v',ineq,n') when Variable.equal v v' && (not (to_fun ineq n n')) -> FalseConstraint
  |AtomConstraint(v,Eq,n),AtomConstraint(v',ineq,n') when Variable.equal v v'                             -> AtomConstraint(v,Eq,n)
  |AtomConstraint(v,Le,n),AtomConstraint(v',Le,n')   when Variable.equal v v'                             -> AtomConstraint(v,Le,min n n')
  |AtomConstraint(v,Gt,n),AtomConstraint(v',Gt,n')   when Variable.equal v v'                             -> AtomConstraint(v,Gt,max n n')
  |AtomConstraint(v,Le,n),AtomConstraint(v',Gt,n')   when Variable.equal v v' && n<=n'                    -> FalseConstraint
  |AtomConstraint(v,Gt,n),AtomConstraint(v',Le,n')   when Variable.equal v v' && n>=n'                    -> FalseConstraint
  |AtomConstraint(v,Le,n),AtomConstraint(v',Gt,n')   when Variable.equal v v' && n=n'+1                   -> AtomConstraint(v,Eq,n)
  |AtomConstraint(v,Gt,n),AtomConstraint(v',Le,n')   when Variable.equal v v' && n'=n+1                   -> AtomConstraint(v,Eq,n')
  |_                                                                                       -> raise UnsupportedConstraint

(* Consequence relation between those constraints. *)
let entails hyp con =
  match hyp,con with
  |FalseConstraint,_
  |_,TrueConstraint                                                           -> true
  |TrueConstraint,_
  |_,FalseConstraint                                                          -> false
  |AtomConstraint(v,Eq,n),AtomConstraint(v',ineq,n') when Variable.equal v v' -> to_fun ineq n n'
  |AtomConstraint(v,Le,n),AtomConstraint(v',Le,n')   when Variable.equal v v' -> n <= n'
  |AtomConstraint(v,Gt,n),AtomConstraint(v',Gt,n')   when Variable.equal v v' -> n >= n'
  |_                                                                          -> false

(* Some predefined relations. *)
let always_greater   ?(hyp=TrueConstraint) t1 t2 = entails hyp (constraint_atom t1 Gt t2)

let always_different ?(hyp=TrueConstraint) t1 t2 =
  match (constraint_atom t1 Eq t2) with
  |TrueConstraint         -> false
  |FalseConstraint        -> true
  |AtomConstraint(v,Eq,n) -> 
      (match hyp with
      |FalseConstraint                                   -> true
      |TrueConstraint                                    -> false
      |AtomConstraint(v',Le,n') when Variable.equal v v' -> n'< n
      |AtomConstraint(v',Gt,n') when Variable.equal v v' -> n'>=n
      |AtomConstraint(v',Eq,n') when Variable.equal v v' -> n'<>n
      |_                                                 -> raise UnsupportedConstraint)
  |_                      -> raise UnsupportedConstraint

(* Normalization for closure
 * Roughly: for closure what we want to know is if two indexes are *always equal* to each other.
 * This depends on the constraints.
 * But the only kind of constraint that can matter is an equality constraint.
 * In which case a trivial normal form can be obtained by just subtituting the parameter by its value in the constraint.
 *
 * Having a normal form allows to use classical data structures to test efficiently if a literal occurs in the structure.
 *)
let normalize_index cstr ~idx =
  match cstr with
  |AtomConstraint(var,Eq,n) -> Indexes.substitute idx ~var ~by:(Indexes.of_num n)
  |_                        -> idx

