(*
 * Handling of indexes e.g. "i+1" in "P_{i+1}", also used for bounds as "1" and "n" in "/\i=1..n ..."
 *
 * Very optimized for regular schemas.
 * Indeed the only authorized expressions are the ones of the form "d.n+k" where:
 * - d = 0 or 1
 * - n is a variable
 * - k is an integer
 *
 * This simplifies much much all the concerned arithmetics (see constraint.ml)
 *)
open Softcore open Looping

module Indexes =
  struct
    (* An index is either an integer or the sum of a variable and integer *)
    type t =
      |Int  of int
      |Plus of Variable.t * int

    (* Constructors *)
    let of_num n = Int n
    let of_var v = Plus(v,0)
    let add t n =
      match t with
      |Int(n')    -> Int(n+n')
      |Plus(v,n') -> Plus(v,n+n')

    (* Selectors *)
    let get_var = function
      |Int(_)    -> None
      |Plus(v,_) -> Some v
    let get_shift = function Int(n) | Plus(_,n) -> n


    (* Printing stuff *)
    let to_parenthesized ~f:to_string s =
      match s with
      |Plus(_,_)  -> "(" ^ (to_string s) ^ ")"
      |_          -> to_string s

    let to_string = function
      |Int(n)             -> string_of_int n
      |Plus(v,n) when n=0 -> (Variable.to_string v)
      |Plus(v,n)          -> (Variable.to_string v) ^ (if n >= 0 then "+" else "-") ^ (string_of_int (abs n))

    let to_tex     = to_string

    (* See substitution.ml *)
    type substituable = t
    type term         = t
    type variable     = Variable.t

    (* Usual substitution of a variable in a first order term *)
    let substitute t ~var:v ~by:t' =
      match t,t' with
      |Plus(v',n), Int(n')      when Variable.equal v' v -> Int(n+n')
      |Plus(v',n), Plus(v'',n') when Variable.equal v' v -> Plus(v'',n+n')
      |t,_                                               -> t

    (* A simple form of substitution that doesn't care which variable is involved. *)
    let compute_term t n =
      match t with
      |Plus(_,n') -> n+n'
      |Int(n')    -> n'

    exception NotInvertible
    let inverse = function
      |Plus(v,k) -> Plus(v,-k)
      |Int _     -> raise NotInvertible

    (* Partial order on expressions:
     * e.g. n < n+1, n-1 < n, 1 < 2
     * but 1 < n is not comparable (n may very well equal 0,1,2 or any other integer)
     *)
    type partially_comparable = Less | Equal | Greater | Incomparable
    exception NotComparable of t * t

    let compare t1 t2 =
      match (t1,t2) with
      |Plus(v1,k1), Plus(v2,k2) when Variable.equal v1 v2 -> if k1 > k2 then Greater else (if k1=k2 then Equal else Less   )
      |Int(n1),Int(n2)                                    -> if n1 > n2 then Greater else (if n1=n2 then Equal else Less   )
      |_                                                  -> Incomparable

    let max t1 t2 =
      match compare t1 t2 with
      |Less         -> t2
      |Greater      -> t1
      |Equal        -> t1
      |Incomparable -> raise (NotComparable(t1,t2))

    let min t1 t2 =
      match compare t1 t2 with
      |Less         -> t1
      |Greater      -> t2
      |Equal        -> t2
      |Incomparable -> raise (NotComparable(t1,t2))

    let total_compare t1 t2 =
      match t1 with
      |Int(n1) ->
          (match t2 with Int(n2) -> n1-n2 |Plus(_,_) -> -1)
      |Plus(v1,k1) ->
          (match t2 with
          |Int _       -> 1
          |Plus(v2,k2) -> let vcmp = Variable.compare v1 v2 in if vcmp <> 0 then vcmp else k1-k2)

    let equal t1 t2 =
      match t1,t2 with
      |Int n1,Int n2 when n1=n2                                   -> true
      |Plus(v1,k1),Plus(v2,k2) when Variable.equal v1 v2 && k1=k2 -> true
      |_                                                          -> false
  end

include Indexes

module LoopingMap =
  struct
    type loopable = Indexes.t
    include Caml.Map.Make(struct type t = Indexes.t let compare x y = total_compare x y end)
    let add m ~key ~data = add key data m
    let find m ~key = try Some (find key m) with Not_found -> None
  end
