(**
 * Indexes.
 *
 * e.g. "i+1" in "P_i+1", also used for bounds as "1" and "n" in "/\i=1..n ..."
 *
 * The only authorized expressions are the ones of the form "d.n+k" where:
 * - d = 0 or 1
 * - n is a variable
 * - k is an integer
 *
 * @author Vincent Aravantinos
 *)

(** The type of indexes *)
type t

(** {6 Constructors } *)

(** An index which is just a number. *)
val of_num : int -> t

(** An index which is just a variable. *)
val of_var : Variable.t -> t

(** The sum of an index and a number. *)
val add    : t -> int -> t

(** {6 Selectors } *)

  (** @return Some v if the index contains a variable, None otherwise *)
val get_var   : t -> Variable.t option
val get_shift : t -> int

(** {6 Printing } *)

val to_string : t -> string
val to_tex    : t -> string

(** {6 Comparisons } *)

(** Computes the maximum of two terms w.r.t. the partial ordering on expressions.
 * e.g. n < n+1, n-1 < n, 1 < 2
 * but 1 < n is not comparable (n may very well equal 0,1,2 or any other integer)
 *)
val max           : t -> t -> t
val min           : t -> t -> t
val equal         : t -> t -> bool
val total_compare : t -> t -> int

(** {6 Substitutions } *)

(** Computes the integer obtained from the substitution of the only variable of t (if any) by the input integer *)
val compute_term : t -> int -> int

include Substitution.Substituable with type substituable = t and type term = t and type variable = Variable.t

(** {6 Looping } *)
module LoopingMap : Looping.LoopingMap with type loopable = t

(** {6 Others } *)
val inverse : t -> t
