(** Looping interface.
 *
 * Handling looping is similar at various scales in the data structures:
 * literals, indexes, schemata, blocks.
 * Hence the following interface.
 *
 * ex. of looping schemata:
 * /\i=1..n-1 P_i loops on /\i=1..n P_i (i.e. they are equivalent up to n).
 *
 * The notion extends similarly to blocks.
 *
 * The current implementation finally uses classical data structures for handling sets of data.
 * The original difficulty with classical data structures is that those allow to find data which is {e equal} to another data in a set
 * whereas we do not want equality but equality {e up to} an equivalence.
 * The first implementations implemented this equivalence hard-wired and usual data structures could not be used.
 * Now we just put every inserted data in a normal form so that we can easily use classical data structures.
 * So the interface is now just a particular subset of usual operations on classical data structures.
 * 
 * The final data structure is a particular kind of trie.
 * So every sub-data involved in a block (schema, literal, index) has to provide a way of mapping its objects to other data.
 * Those maps are used to construct the children of a node in the trie.
 *
 * @see <http://membres-liglab.imag.fr/aravantinos/Publis/tableaux2009/tab09.pdf> A Schemata Calculus For Propositional Logic.
 * @see <http://en.wikipedia.org/wiki/Trie> Trie (Wikipedia)
 *
 * @author Vincent Aravantinos
 *)

(*
 * The interface for a map.
 * Called "LoopingMap" for legacy, there's now nothing more than just a usual map.
 *)
module type LoopingMap =
  sig
    (** Type of the ``relevant'' data. *)
    type loopable

    (** Type of a set of loopable datas. *)
    type 'a t

    val empty : 'a t
    val add   : 'a t -> key:loopable -> data: 'a -> 'a t
    val find  : 'a t -> key:loopable -> 'a option
  end
