open Softcore open Lexing open Schema

let non_regular_string       = "Error: the input schema is not regular, "
let more_than_one_string     = "more than one parameter was detected, bad parenthesing?"
let diff_lower_bounds_string = "two iterations with different lower bounds were detected."
let inside_iteration_string  = "the parameter occurs inside an iteration."

(* The main function, called with the input channel *)
let main ~verbose ?fname ch =
  if ch = stdin then print_endline "Taking input from stdin.";
  let lexbuf = 
    let lexbuf' = from_channel ch in
    match fname with None -> lexbuf' | Some fn -> { lexbuf' with lex_curr_p = { lexbuf'.lex_curr_p with pos_fname = fn } } in
  print_endline 
  (try
    let (sch,cstr) = Parser.main Lexer.token lexbuf in
    (if verbose then (Schema.to_string sch) ^ "\n" else "")
    ^
    (match (Schema.is_regular sch) with
    |`No argument -> non_regular_string ^ "\n" ^
      (match argument with
      |Different_lower_bounds -> diff_lower_bounds_string
      |More_than_one_param    -> more_than_one_string
      |Param_inside_iteration -> inside_iteration_string
      )
    |`Yes ->
        let result = Regstab.regstab ~cstr sch in
        if Satisfiability.satisfiable result then "satisfiable" ^
          (match Satisfiability.model_hint result with 
          |None                                -> ""
          |Some c when Constraint.equal c cstr -> ""
          |Some c                              -> " (when " ^ Constraint.to_string c ^ ")")
        else "unsatisfiable")
  with Failure s -> s
  );
  print_newline ();
  flush stdout

(* Handling of command-line options, calls main *)
let _ =
  let channel = ref None in
  let verbose = ref false in
  let options = [
    "-verbose", Arg.Set(verbose), "Be verbose (i.e. currently, print the input schema after parsing)"
  ] in
  let usage = "regstab [file]    - says if the schema in file is satisfiable, schema taken on stdin if no file provided" in
  let file s = channel := Some(s) in
  Arg.parse options file usage;
  match !channel with
  |None      -> main stdin ~verbose:(!verbose)
  |Some file -> read_wrap ~f:(main ~verbose:(!verbose) ~fname:file) file

