open Softcore open Block open Schema open Constraint open Satisfiability

(*
 * The following global (imperative) variable contains a set of blocks that have already been seen, called ``lemmas''.
 * This set will be inspected for looping.
 *)
let lemmas = Lemmas.empty

(*
 * Propositional rules apply as long as possible.
 * As soon as a schema is encountered that can be applied a rule, it is decomposed according to this rule.
 * prop_rules is then applied recursively on the new nodes (the recursion is done through continuations).
 * Closure rule is also applied as soon as possible.
 *)
let rec prop_rules bck = 
  Block.apply_rules_with_closure bck ~k:pure_rule ~rules:(fun ~sch ~cstr ->
    match sch with
    |(`Op(`And,sch1,sch2)   : Schema.t) -> Update(Replace_with_schemas(sch1,sch2))
    |(`Op(`Or,sch1,sch2)    : Schema.t) -> Branch(Replace_with_schema sch1, Replace_with_schema sch2)
    |(`Op(`Xor,sch1,sch2)   : Schema.t) -> Branch(Replace_with_schemas(neg sch1,sch2), Replace_with_schemas(sch1,neg sch2))
    |(`Op(`Equiv,sch1,sch2) : Schema.t) -> Branch(Replace_with_schemas(neg sch1,neg sch2), Replace_with_schemas(sch1,sch2))
    |(`Op(`Imply,sch1,sch2) : Schema.t) -> Branch(Replace_with_schema(neg sch1), Replace_with_schema(sch2))
    |(`Bot                  : Schema.t) -> Close
    |_                                  -> Go_on)

and pure_rule bck =
  let bck' = Block.dedup bck in
  match Block.purify bck' ~cstr:(Block.cstr_of bck') ~pur:(Block.purity bck') with
  |Modified.Unmodified,bck'   -> trivial_rule bck'
  |Modified.Modified,purified -> trivial_rule purified

and trivial_rule bck =
  if Block.is_empty bck then ModelFound (cstr_of bck) else only_literals_rule bck

(* If a block contains only literals and closure rule did not apply then it is satisfiable
 * (not proved anywhere but trivial)
 *)
and only_literals_rule bck = if Block.contains_only_literals bck then ModelFound (cstr_of bck) else looping_rule bck

(* The looping rule checks if the current node loops on a previously seen node (``lemma''), in which case we can conclude (NoModelFound) *)
and looping_rule bck =
  if Block.contains_iterations bck
  then (if Lemmas.loops_among lemmas bck then NoModelFound else add_lemma bck)
  else add_lemma bck

and add_lemma bck = Lemmas.add lemmas bck; it_rules bck

(* Iteration rules apply in a way similar to propositional rules *)
and it_rules bck =
  match Block.max_bound bck with
  |None     -> constraint_rule bck
  |Some max -> Block.apply_rules bck ~k:constraint_rule ~rules:(fun ~sch ~cstr ->
      match sch with
      |(`It(`And,var,t1,t2,sch) : Schema.t) when Indexes.equal t2 max (* should always be the case except with relaxed conditions *) ->
          Branch(Replace_with_schema_add_constraint(unfold_and_once ~var ~sch ~t1 ~t2, bounds_constraint t1 Le t2),
            Add_constraint(bounds_constraint t1 Gt t2))
      |(`It(`Or,var,t1,t2,sch) : Schema.t) when Indexes.equal t2 max  ->
          Update(Replace_with_schema_add_constraint(unfold_or_once ~var ~sch ~t1 ~t2, bounds_constraint t1 Le t2))
      |(`Bot : Schema.t)                                              -> Close
      |_                                                              -> Go_on)

(* The constraint rule closes the tableau if the current constraint is unsatisfiable. *)
and constraint_rule bck = if Constraint.is_false (Block.cstr_of bck) then NoModelFound else prop_rules bck

and regstab ?(cstr=Constraint.top) sch = prop_rules (Block.of_schs_and_cstr ~schs:[sch] ~cstr)
