(* Schemata. *)
open Softcore open Looping

exception Should_not_occur of string

module Schema = 
  struct
    type op = [ `And | `Or ]
    type ext_op = [ `And | `Or | `Xor | `Equiv | `Imply ]
    (* A schema is either true, or false or a literal or the application of a usual, propositional, connective ('`Op')
     * or the iteration of such a connective ('`It').
     * 'op' is the type of connectives.
     *)
    type t =
      [`Lit of Literal.t
      |`Op  of ext_op * t * t
      |`It  of op * Variable.t * Indexes.t * Indexes.t * t
      |`Top
      |`Bot ]

    (* When a schema is not regular *)
    exception Non_regular

    (* A schema is loopable and substituable *)
    type substituable = t
    type term         = Indexes.t
    type variable     = Variable.t

    (* Constructors *)
    let bot                   = `Bot
    let top                   = `Top
    let pos_lit prop idx      = `Lit(Literal.pos prop idx)
    let neg_lit prop idx      = `Lit(Literal.neg prop idx)
    let conj sch1 sch2        = `Op(`And,sch1,sch2)
    let disj sch1 sch2        = `Op(`Or,sch1,sch2)
    let xor sch1 sch2         = `Op(`Xor,sch1,sch2)
    let equiv sch1 sch2       = `Op(`Equiv,sch1,sch2)
    let imply sch1 sch2       = `Op(`Imply,sch1,sch2)
    let it_conj var t1 t2 sch = `It(`And,var,t1,t2,sch)
    let it_disj var t1 t2 sch = `It(`Or,var,t1,t2,sch)

    (* Neutral element of a connective. *)
    let neutral = function `And -> `Top | `Or -> `Bot

    (* Print stuff *)

    let to_parenthesized ~f:to_string s = 
      match s with
      |`Lit(_)
      |`Top
      |`Bot         -> to_string s
      |_           -> "(" ^ (to_string s) ^ ")"

    let op_to_string = function `And -> "/\\" | `Or -> "\\/" | `Xor -> "(+)" | `Equiv -> "<->" | `Imply -> "->"

    let op_equal op1 op2 =
      match op1,op2 with
      | `And,`And | `Or,`Or | `Xor,`Xor | `Equiv,`Equiv | `Imply,`Imply -> true
      | _                                                               -> false

    let op_compare op1 op2 =
      match op1,op2 with
      (*FIXME: no wildcard *)
      | `And,`And | `Or,`Or | `Xor,`Xor | `Equiv,`Equiv | `Imply,`Imply -> 0
      | `And,_                                                          -> -1
      | _,`And                                                          -> 1
      | `Or,_                                                           -> -1
      | _,`Or                                                           -> 1
      | `Xor,_                                                          -> -1
      | _,`Xor                                                          -> 1
      | `Equiv,_                                                        -> -1
      | _,`Equiv                                                        -> 1

    let rec to_string = function
      |`Top              -> "True"
      |`Bot              -> "False"
      |`Lit(l)           -> Literal.to_string l
      |`Op(op,s1,s2)     -> (to_parenthesized ~f:to_string s1) ^ " " ^ (op_to_string op) ^ " " ^ (to_parenthesized ~f:to_string s2)
      |`It(op,i,t1,t2,s) -> (op_to_string op) ^ (Variable.to_string i) ^ "=" ^ (Indexes.to_string t1) ^ ".." ^ (Indexes.to_string t2)
                              ^ " " ^ (to_parenthesized ~f:to_string s)
    let list_to_string schs = 
      let rec list_to_string' = function
        |[] -> ""
        |[x] -> to_string x
        |x::xs -> (to_string x) ^ " ; " ^ (list_to_string' xs)
      in
      "[" ^ (list_to_string' schs) ^ "]"

    let print = print_endline <|| to_string

    (* Substitution. Just propagate till the literals. *)
    let rec substitute t ~var:v ~by:t' = 
      match t with
        |`Lit(l)                                           -> `Lit(Literal.substitute ~var:v ~by:t' l)
        |`Op(op,s1,s2)                                     -> `Op(op,substitute ~var:v ~by:t' s1,substitute ~var:v ~by:t' s2)
        |`It(op,v',t1,t2,s) when not (Variable.equal v' v) -> `It(op,v',t1,Indexes.substitute ~var:v ~by:t' t2,substitute ~var:v ~by:t' s)
        |`It(_,_,_,_,_) as s                               -> s
        |x                                                 -> x

    (* Get the parameters.
     * We use a Set because we don't duplicate entries. *)
    let rec get_params ?(acc=Set.empty) = function
      |`Lit(l)          -> (match Indexes.get_var (Literal.idx_of l) with None -> acc | Some v -> Set.add acc v)
      |`Op(_,sch1,sch2) -> get_params ~acc:(get_params ~acc sch1) sch2
      |`It(_,_,_,t,_)   -> (match Indexes.get_var t with None -> acc | Some v -> Set.add acc v)
      |_               -> acc

    (* Has the schema more than 1 param?
     *)
    let has_many_params sch = Set.cardinal (get_params sch) > 1

    (* Get *the* parameter, if any. Raises Non_regular if more than 1.
     *)
    let get_param sch = if has_many_params sch then raise Non_regular else Set.choose (get_params sch)

    (* Trivial conversion of schemata ops to formula ops *)
    let to_formula = function `And -> Formula.conj | `Or -> Formula.disj | `Xor -> Formula.xor | `Equiv -> Formula.equiv | `Imply -> Formula.imply

    (* A simple form of substitution that doesn't care which variable is involved. 
     * Generates a formula i.e. a data of type Formula.t
     *)
    let rec compute_schema sch n =
      if Set.cardinal (get_params sch) <= 1 then
        match sch with
        |`Lit(l)                -> Formula.lit   (Literal.compute_lit l n)
        |`Op(op,sch1,sch2)      -> (to_formula op) (compute_schema sch1 n) (compute_schema sch2 n)
        |`It(op,var,t1,t2,sch') -> 
            (match List.map ~f:(compute_schema sch') (List.range (Indexes.compute_term t1 n) ((Indexes.compute_term t2 n)+1)) with
            |[]     -> (match op with `And -> Formula.top | `Or -> Formula.bot)
            |[rank] -> rank
            |ranks  -> List.reduce ~f:(to_formula op) ranks)
        |`Top                   -> Formula.top
        |`Bot                   -> Formula.bot
      else raise Non_regular 

    (* Computes the maximal of upper bounds occurring in some schema, if there are iterations. *)
    let rec max_bound ?(acc=None) sch =
      match sch,acc with
      |`Op(_,sch1,sch2),_        -> max_bound ~acc:(max_bound ~acc sch1) sch2
      |`It(_,_,_,t2,_),None      -> Some(t2)
      |`It(_,_,_,t2,_),Some(t2') -> Some(Indexes.max t2 t2')
      |_,_                      -> acc

    (* True iff the schema contains iterations *)
    let rec contains_iterations = function
      |`Op(_,sch1,sch2) -> (contains_iterations sch1) || (contains_iterations sch2)
      |`It(_,_,_,_,_)   -> true
      |_               -> false

    (* Regularity checks. A schema is regular if:
     * - no iterations are nested                                                           (ensured by parsing)
     * - indexes of iterations are of the form i+k where i is a variable, k an integer      (ensured by parsing)
     * - indexed propositions are monadic i.e. only one index                               (ensured by parsing)
     * - lower bounds of iterations must be integers                                        (ensured by parsing)
     * - upper bounds must be of the form n+k, where n is the parameter and k is an integer (ensured by parsing && def of parameter)
     * - iterations must be aligned i.e. same bounds for all iterations                     
     *    -> RELAXED, except for lower bounds, RegSTAB still terminates though not proved in the paper
     * - it has only one parameter                                                          (done by get_param)
     * - in the previous expression i+k, i cannot be the parameter                          (done by variable_occurs_inside_iteration)
     *)

    let variable_occurs_inside_iteration sch ~var =
      let rec variable_occurs_inside_iteration' ~acc ~inside_an_iteration = function
        |`Lit(l) when inside_an_iteration ->
            (match Indexes.get_var (Literal.idx_of l) with 
            |Some(v) -> acc || Variable.equal v var
            |None    -> acc)
        |`Op(_,sch1,sch2) -> variable_occurs_inside_iteration' sch2 ~acc:(variable_occurs_inside_iteration' sch1 ~acc ~inside_an_iteration) ~inside_an_iteration
        |`It(_,_,_,_,sch) -> variable_occurs_inside_iteration' sch  ~acc ~inside_an_iteration:true
        |_ -> acc
      in
      variable_occurs_inside_iteration' ~acc:false ~inside_an_iteration:false sch

    let same_lower_bounds sch =
      let rec lower_bounds = function
        |`Op(_,sch1,sch2) -> Set.union (lower_bounds sch1) (lower_bounds sch2)
        |`It(_,_,t1,_,_)  -> Set.singleton t1
        |_               -> Set.empty
      in
      (Set.cardinal (lower_bounds sch)) <= 1
        
    type non_regular_arg = Different_lower_bounds | More_than_one_param | Param_inside_iteration
    let is_regular sch =
      if not (same_lower_bounds sch) then `No Different_lower_bounds
      else 
        let params = get_params sch in
        match Set.cardinal params with
        |0                                                                        -> `Yes
        |1 when variable_occurs_inside_iteration sch ~var:(Set.choose_exn params) -> `No Param_inside_iteration
        |1                                                                        -> `Yes
        |_                                                                        -> `No More_than_one_param

    (* Input: only literals or iterations.
     * This function is intended to be used only with lemmas.
     *)
    let expr_involving_param sch =
      match get_param sch with
      |None   -> None
      |Some _ -> 
          let involve_var_or_nothing t = if Option.is_none (Indexes.get_var t) then None else Some t in
          (match sch with
          |`Lit l          -> involve_var_or_nothing (Literal.idx_of l)
          |`It(op,_,_,t,_) -> involve_var_or_nothing t
          |_              -> raise(Should_not_occur("expr_involving_param called with a schema which is neither an iteration nor a literal"))
          )

    (* Computes the negation in negative normal form of a whole schema. *)
    let rec neg = function
      |`Top                  -> `Bot
      |`Bot                  -> `Top
      |`Lit(l)               -> `Lit(Literal.opposite l)
      |`Op(op, sch1, sch2)   -> neg_op op sch1 sch2
      |`It(`And,i,t1,t2,sch) -> `It(`Or,i,t1,t2,neg sch)
      |`It(`Or,i,t1,t2,sch)  -> `It(`And,i,t1,t2,neg sch)

    and neg_op op sch1 sch2 = 
      match op with
        |`And   -> `Op(`Or   , neg sch1, neg sch2)
        |`Or    -> `Op(`And  , neg sch1, neg sch2)
        |`Xor   -> `Op(`Equiv, sch1, sch2)
        |`Equiv -> `Op(`Xor  , sch1, sch2)
        |`Imply -> `Op(`And  , sch1, neg sch2)

    (* Symbolically unfold the last rank of iterations.
     * I.e. /\i=a..b S -> (/\i=a..b-1 S) /\ S{i<-b} 
     *)
    let unfold_once ~op ~var ~sch ~t1 ~t2 = `Op((op:>ext_op),substitute ~var ~by:t2 sch, `It(op,var, t1, Indexes.add t2 (-1), sch))
    let unfold_or_once ~var ~sch ~t1 ~t2 = unfold_once ~op:`Or ~var ~sch ~t1 ~t2 
    let unfold_and_once ~var ~sch ~t1 ~t2 = unfold_once ~op:`And ~var ~sch ~t1 ~t2 

    type purable = t

    (* Look for literals in the schema, and add them to the purity. *)
    let purity_add sch ?(range=None) ~pur =
      let neg_sgn = function `Positive -> `Negative | `Negative -> `Positive | `Both -> `Both in
      let rec purity_add sch ?(range=None) ~pur ~sgn =
        match sch with
        |`Op(`Xor,sch1,sch2)
        |`Op(`Equiv,sch1,sch2) -> purity_add ~range ~pur:(purity_add ~range ~pur ~sgn:`Both sch1) ~sgn:`Both sch2
        |`Op(`Imply,sch1,sch2) -> purity_add ~range ~pur:(purity_add ~range ~pur ~sgn:(neg_sgn sgn) sch1) ~sgn sch2
        |`Op(_,sch1,sch2)      -> purity_add ~range ~pur:(purity_add ~range ~pur ~sgn sch1) ~sgn sch2
        |`It(_,var,t1,t2,sch)  -> purity_add ~pur ~range:(Some(var,t1,t2)) ~sgn sch
        |`Lit(l)              -> 
            (match sgn with
            |`Positive -> Purity.add ~range ~lit:l pur
            |`Negative -> Purity.add ~range ~lit:(Literal.opposite l) pur
            |`Both     -> Purity.add ~range ~lit:l (Purity.add ~range ~lit:(Literal.opposite l) pur))
        |_                   -> pur
      in
      purity_add sch ~range ~pur ~sgn:`Positive

    (* When the schema is a pure literal return None, keep it as is otherwise.
     * Notice we don't dig inside the schema to look for pure literals.
     * That's the job of rules to decompose the schema in order to get only literals.
     *)
    let purify sch ~cstr ~pur =
      match sch with
      |`Lit l when Purity.is_pure ~cstr ~pur l -> Modified.Modified, `Top
      |_                                      -> Modified.Unmodified, sch

    (* Comparisons *)
    let rec total_compare s1 s2 =
      match s1,s2 with
      |`Lit l1, `Lit l2                -> Literal.compare l1 l2
      |`Op(op1,s1,s1'),`Op(op2,s2,s2') -> 
          let opcmp = op_compare op1 op2 in 
          if opcmp <> 0 then opcmp else
            (let scmp = total_compare s1 s2 in
            if scmp <> 0 then scmp else total_compare s1' s2')
      |`It(op1,v1,_,t1,s1),`It(op2,v2,_,t2,s2) ->
          let opcmp = op_compare op1 op2 in 
          if opcmp <> 0 then opcmp else
            (let vcmp = Variable.compare v1 v2 in
            if vcmp <> 0 then vcmp else 
              (let tcmp = Indexes.total_compare t1 t2 in
              if tcmp <> 0 then tcmp else total_compare s1 s2)
            )
      |`Top,`Top | `Bot,`Bot -> 0
      |`Bot,_             -> -1
      |_,`Bot             -> 1
      |`Lit _,_           -> -1
      |_,`Lit _           -> 1
      |`Op _,_            -> -1
      |_,`Op _            -> 1
      |`It _,_            -> -1
      |_,`It _            -> 1

    let rec equal s1 s2 =
      match s1,s2 with
      |`Top,`Top | `Bot,`Bot                         -> true
      |`Lit l1, `Lit l2                            -> Literal.equal l1 l2
      |`Op(op1,s1,s1'),`Op(op2,s2,s2')             -> op_equal op1 op2 && equal s1 s2 && equal s1' s2'
      |`It(op1,v1,t1,t1',s1),`It(op2,v2,t2,t2',s2) -> 
          op_equal (op1:>ext_op) (op2:>ext_op) && Variable.equal v1 v2 && Indexes.equal t1 t2 && Indexes.equal t1' t2' && equal s1 s2
      |_                                         -> false

    (* Substitution of propositional variables *)
    module PropSubs : Substitution.Substituable with type substituable = t and type term = Proposition.t and type variable = Proposition.t =
      struct
        type substituable = t
        type term         = Proposition.t
        type variable     = Proposition.t
        (* Substitution. Just propoagate till the literals. *)
        let rec substitute t ~var:p ~by:p' =
          match t with
            |`Lit(l)             -> `Lit(Literal.PropSubs.substitute ~var:p ~by:p' l)
            |`Op(op,s1,s2)       -> `Op(op,substitute ~var:p ~by:p' s1,substitute ~var:p ~by:p' s2)
            |`It(op,v,t1,t2,s)   -> `It(op,v,t1,t2,substitute ~var:p ~by:p' s)
            |x                  -> x
      end
  end

include Schema


(* All the following is to define a map on schemata which may be inserted in the set of lemmas.
 *)

(* A flat schema can be handled without caring about the shift as the parameter does not occur inside it. *)
module FlatSchemaMap : Looping.LoopingMap with type loopable = Schema.t =
  struct
    type loopable = Schema.t
    include Caml.Map.Make(struct type t = loopable let compare x y = total_compare x y end)
    let add m ~key ~data = add key data m
    let find m ~key = try Some (find key m) with Not_found -> None
  end

module IterationMap : Looping.LoopingMap with type loopable = Schema.t =
  struct
    type loopable = Schema.t
    module OrderedOpVarIdx =
      struct
        type t = op * Variable.t * Indexes.t
        let compare (op1,v1,t1) (op2,v2,t2) =
          let opcmp = op_compare op1 op2 in
          if opcmp <> 0 then opcmp else
            (let vcmp = Variable.compare v1 v2 in
            if vcmp <> 0 then vcmp else Indexes.total_compare t1 t2)
      end
    module M = Caml.Map.Make(OrderedOpVarIdx)
    type 'a t = ('a FlatSchemaMap.t) M.t
    let empty = M.empty
    let add m ~key ~data =
      match key with
      |`It(op,v,_,t,s) -> M.add (op,v,t) (FlatSchemaMap.add (try M.find (op,v,t) m with Not_found -> FlatSchemaMap.empty) ~key:s ~data) m
      |_              -> raise (Should_not_occur("IterationMap taking a non iteration key"))
    let find m ~key =
      match key with
      |`It(op,v,_,t2,s) -> (try FlatSchemaMap.find (M.find (op,v,t2) m) ~key:s with Not_found -> None)
      |_               -> raise (Should_not_occur("IterationMap taking a non iteration key"))
  end

module LoopingMap =
  struct
    module LitMap = Literal.LoopingMap
    type loopable = Schema.t
    type 'a t = 'a LitMap.t * 'a IterationMap.t * 'a option * 'a option
    let empty = (LitMap.empty,IterationMap.empty,None,None)
    let add (litm,itm,bot,top) ~key ~data = 
      match key with
      |`Lit l         -> (LitMap.add litm ~key:l ~data,itm,bot,top)
      |`It(_,_,_,_,_) -> (litm,IterationMap.add itm ~key ~data,bot,top)
      |`Bot           -> (litm,itm,Some data,top)
      |`Top           -> (litm,itm,bot,Some data)
      |_             -> raise (Should_not_occur("A lemma inserted in the lemmas base should not contain propositional stuff"))
    let find (litm,itm,bot,top) ~key =
      match key with
      |`Lit l         -> LitMap.find litm ~key:l
      |`It(_,_,_,_,_) -> IterationMap.find itm ~key
      |`Bot           -> bot
      |`Top           -> top
      |_             -> raise (Should_not_occur("A lemma searched in the lemmas base should not contain propositional stuff"))
  end
